@@ -26,6 +26,7 @@ import sys
 import copy
 import time
 import traceback
+from typing import Dict, List
 from firewall import config
 from firewall import functions
 from firewall.core import ipXtables
@@ -42,8 +43,10 @@ from firewall.core.fw_policies import FirewallPolicies
 from firewall.core.fw_ipset import FirewallIPSet
 from firewall.core.fw_transaction import FirewallTransaction
 from firewall.core.fw_helper import FirewallHelper
+from firewall.core.fw_policy import FirewallPolicy
 from firewall.core.fw_nm import nm_get_bus_name, nm_get_interfaces_in_zone
 from firewall.core.logger import log
+from firewall.core.io.io_object import IO_Object
 from firewall.core.io.firewalld_conf import firewalld_conf
 from firewall.core.io.direct import Direct
 from firewall.core.io.service import service_reader
@@ -52,6 +55,8 @@ from firewall.core.io.zone import zone_reader, Zone
 from firewall.core.io.ipset import ipset_reader
 from firewall.core.ipset import IPSET_TYPES
 from firewall.core.io.helper import helper_reader
+from firewall.core.io.policy import policy_reader
+from firewall.core.rich import Rich_Rule
 from firewall import errors
 from firewall.errors import FirewallError
 
@@ -76,15 +81,15 @@ class Firewall(object):
         else:
             self.ip4tables_backend = ipXtables.ip4tables(self)
             self.ip4tables_enabled = True
-            self.ip4tables_supported_icmp_types = [ ]
+            self.ipv4_supported_icmp_types = [ ]
             self.ip6tables_backend = ipXtables.ip6tables(self)
             self.ip6tables_enabled = True
-            self.ip6tables_supported_icmp_types = [ ]
+            self.ipv6_supported_icmp_types = [ ]
             self.ebtables_backend = ebtables.ebtables()
             self.ebtables_enabled = True
             self.ipset_backend = ipset.ipset()
             self.ipset_enabled = True
-            self.ipset_supported_types = [ ]
+            self.ipset_supported_types = IPSET_TYPES
             self.nftables_backend = nftables.nftables(self)
             self.nftables_enabled = True
 
@@ -98,25 +103,30 @@ class Firewall(object):
         self.policies = FirewallPolicies()
         self.ipset = FirewallIPSet(self)
         self.helper = FirewallHelper(self)
+        self.policy = FirewallPolicy(self)
 
         self.__init_vars()
 
     def __repr__(self):
-        return '%s(%r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r)' % \
+        return '%s(%r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r)' % \
             (self.__class__, self.ip4tables_enabled, self.ip6tables_enabled,
              self.ebtables_enabled, self._state, self._panic,
              self._default_zone, self._module_refcount, self._marks,
-             self.cleanup_on_exit, self.ipv6_rpfilter_enabled,
-             self.ipset_enabled, self._individual_calls, self._log_denied)
+             self.cleanup_on_exit, self.cleanup_modules_on_exit,
+             self.ipv6_rpfilter_enabled, self.ipset_enabled,
+             self._individual_calls, self._log_denied)
 
     def __init_vars(self):
         self._state = "INIT"
         self._panic = False
         self._default_zone = ""
+        self._default_zone_interfaces = []
+        self._nm_assigned_interfaces = []
         self._module_refcount = { }
         self._marks = [ ]
         # fallback settings will be overloaded by firewalld.conf
         self.cleanup_on_exit = config.FALLBACK_CLEANUP_ON_EXIT
+        self.cleanup_modules_on_exit = config.FALLBACK_CLEANUP_MODULES_ON_EXIT
         self.ipv6_rpfilter_enabled = config.FALLBACK_IPV6_RPFILTER
         self._individual_calls = config.FALLBACK_INDIVIDUAL_CALLS
         self._log_denied = config.FALLBACK_LOG_DENIED
@@ -125,24 +135,74 @@ class Firewall(object):
         self._rfc3964_ipv4 = config.FALLBACK_RFC3964_IPV4
         self._allow_zone_drifting = config.FALLBACK_ALLOW_ZONE_DRIFTING
 
-    def individual_calls(self):
-        return self._individual_calls
+    def get_all_io_objects_dict(self):
+        """
+        Returns a dict of dicts of all runtime config objects.
+        """
+        conf_dict = {}
+        conf_dict["ipsets"] = {_ipset: self.ipset.get_ipset(_ipset) for _ipset in self.ipset.get_ipsets()}
+        conf_dict["helpers"] = {helper: self.helper.get_helper(helper) for helper in self.helper.get_helpers()}
+        conf_dict["icmptypes"] = {icmptype: self.icmptype.get_icmptype(icmptype) for icmptype in self.icmptype.get_icmptypes()}
+        conf_dict["services"] = {service: self.service.get_service(service) for service in self.service.get_services()}
+        conf_dict["zones"] = {zone: self.zone.get_zone(zone) for zone in self.zone.get_zones()}
+        conf_dict["policies"] = {policy: self.policy.get_policy(policy) for policy in self.policy.get_policies_not_derived_from_zone()}
+
+        # The runtime might not actually support all the defined icmptypes.
+        # This is the case if ipv6 (ip6tables) is disabled. Unfortunately users
+        # disable IPv6 and also expect the IPv6 stuff to be silently ignored.
+        # This is problematic for defaults that include IPv6 stuff, e.g. policy
+        # 'allow-host-ipv6'. Use this to make a better decision about errors vs
+        # warnings.
+        #
+        conf_dict["icmptypes_unsupported"] = {}
+        for icmptype in (set(self.config.get_icmptypes()).difference(
+                         set(self.icmptype.get_icmptypes()))):
+            conf_dict["icmptypes_unsupported"][icmptype] = self.config.get_icmptype(icmptype)
+        # Some icmptypes support multiple families. Add those that are missing
+        # support for a subset of families.
+        for icmptype in (set(self.config.get_icmptypes()).intersection(
+                         set(self.icmptype.get_icmptypes()))):
+            if icmptype not in self.ipv4_supported_icmp_types or \
+               icmptype not in self.ipv6_supported_icmp_types:
+                conf_dict["icmptypes_unsupported"][icmptype] = copy.copy(self.config.get_icmptype(icmptype))
+                conf_dict["icmptypes_unsupported"][icmptype].destination = []
+                if icmptype not in self.ipv4_supported_icmp_types:
+                    conf_dict["icmptypes_unsupported"][icmptype].destination.append("ipv4")
+                if icmptype not in self.ipv6_supported_icmp_types:
+                    conf_dict["icmptypes_unsupported"][icmptype].destination.append("ipv6")
+
+        return conf_dict
+
+    def full_check_config(self, extra_io_objects: Dict[str, List[IO_Object]] = {}):
+        all_io_objects = self.get_all_io_objects_dict()
+        # mix in the extra objects
+        for type_key in extra_io_objects:
+            for obj in extra_io_objects[type_key]:
+                all_io_objects[type_key][obj.name] = obj
+
+        # we need to check in a well defined order because some io_objects will
+        # cross-check others
+        order = ["ipsets", "helpers", "icmptypes", "services", "zones", "policies"]
+        for io_obj_type in order:
+            io_objs = all_io_objects[io_obj_type]
+            for (name, io_obj) in io_objs.items():
+                io_obj.check_config_dict(io_obj.export_config_dict(), all_io_objects)
 
     def _check_tables(self):
         # check if iptables, ip6tables and ebtables are usable, else disable
         if self.ip4tables_enabled and \
-           "filter" not in self.get_backend_by_ipv("ipv4").get_available_tables():
-            log.warning("iptables not usable, disabling IPv4 firewall.")
+           "filter" not in self.ip4tables_backend.get_available_tables():
+            log.info1("iptables is not usable.")
             self.ip4tables_enabled = False
 
         if self.ip6tables_enabled and \
-           "filter" not in self.get_backend_by_ipv("ipv6").get_available_tables():
-            log.warning("ip6tables not usable, disabling IPv6 firewall.")
+           "filter" not in self.ip6tables_backend.get_available_tables():
+            log.info1("ip6tables is not usable.")
             self.ip6tables_enabled = False
 
         if self.ebtables_enabled and \
-           "filter" not in self.get_backend_by_ipv("eb").get_available_tables():
-            log.warning("ebtables not usable, disabling ethernet bridge firewall.")
+           "filter" not in self.ebtables_backend.get_available_tables():
+            log.info1("ebtables is not usable.")
             self.ebtables_enabled = False
 
         # is there at least support for ipv4 or ipv6
@@ -155,10 +215,13 @@ class Firewall(object):
         try:
             self.ipset_backend.set_list()
         except ValueError:
-            log.warning("ipset not usable, disabling ipset usage in firewall.")
-            # ipset is not usable, no supported types
+            if self.nftables_enabled:
+                log.info1("ipset not usable, disabling ipset usage in firewall. Other set backends (nftables) remain usable.")
+            else:
+                log.warning("ipset not usable, disabling ipset usage in firewall.")
+                self.ipset_supported_types = [ ]
+            # ipset is not usable
             self.ipset_enabled = False
-            self.ipset_supported_types = [ ]
         else:
             # ipset is usable, get all supported types
             self.ipset_supported_types = self.ipset_backend.set_supported_types()
@@ -169,36 +232,52 @@ class Firewall(object):
                 log.warning("iptables-restore is missing, using "
                             "individual calls for IPv4 firewall.")
             else:
-                log.warning("iptables-restore and iptables are missing, "
-                            "disabling IPv4 firewall.")
+                if self.nftables_enabled:
+                    log.info1("iptables-restore and iptables are missing, "
+                              "IPv4 direct rules won't be usable.")
+                else:
+                    log.warning("iptables-restore and iptables are missing, "
+                                "disabling IPv4 firewall.")
                 self.ip4tables_enabled = False
-        if self.ip4tables_enabled:
-            self.ip4tables_supported_icmp_types = \
-                self.ip4tables_backend.supported_icmp_types()
+        if self.nftables_enabled:
+            self.ipv4_supported_icmp_types = self.nftables_backend.supported_icmp_types("ipv4")
         else:
-            self.ip4tables_supported_icmp_types = [ ]
+            if self.ip4tables_enabled:
+                self.ipv4_supported_icmp_types = self.ip4tables_backend.supported_icmp_types()
+            else:
+                self.ipv4_supported_icmp_types = [ ]
         self.ip6tables_backend.fill_exists()
         if not self.ip6tables_backend.restore_command_exists:
             if self.ip6tables_backend.command_exists:
                 log.warning("ip6tables-restore is missing, using "
                             "individual calls for IPv6 firewall.")
             else:
-                log.warning("ip6tables-restore and ip6tables are missing, "
-                            "disabling IPv6 firewall.")
+                if self.nftables_enabled:
+                    log.info1("ip6tables-restore and ip6tables are missing, "
+                              "IPv6 direct rules won't be usable.")
+                else:
+                    log.warning("ip6tables-restore and ip6tables are missing, "
+                                "disabling IPv6 firewall.")
                 self.ip6tables_enabled = False
-        if self.ip6tables_enabled:
-            self.ip6tables_supported_icmp_types = \
-                self.ip6tables_backend.supported_icmp_types()
+        if self.nftables_enabled:
+            self.ipv6_supported_icmp_types = self.nftables_backend.supported_icmp_types("ipv6")
         else:
-            self.ip6tables_supported_icmp_types = [ ]
+            if self.ip6tables_enabled:
+                self.ipv6_supported_icmp_types = self.ip6tables_backend.supported_icmp_types()
+            else:
+                self.ipv6_supported_icmp_types = [ ]
         self.ebtables_backend.fill_exists()
         if not self.ebtables_backend.restore_command_exists:
             if self.ebtables_backend.command_exists:
                 log.warning("ebtables-restore is missing, using "
                             "individual calls for bridge firewall.")
             else:
-                log.warning("ebtables-restore and ebtables are missing, "
-                            "disabling bridge firewall.")
+                if self.nftables_enabled:
+                    log.info1("ebtables-restore and ebtables are missing, "
+                              "eb direct rules won't be usable.")
+                else:
+                    log.warning("ebtables-restore and ebtables are missing, "
+                                "disabling bridge firewall.")
                 self.ebtables_enabled = False
 
         if self.ebtables_enabled and not self._individual_calls and \
@@ -228,6 +307,13 @@ class Firewall(object):
                 log.debug1("CleanupOnExit is set to '%s'",
                            self.cleanup_on_exit)
 
+            if self._firewalld_conf.get("CleanupModulesOnExit"):
+                value = self._firewalld_conf.get("CleanupModulesOnExit")
+                if value is not None and value.lower() in [ "yes", "true" ]:
+                    self.cleanup_modules_on_exit = True
+                log.debug1("CleanupModulesOnExit is set to '%s'",
+                           self.cleanup_modules_on_exit)
+
             if self._firewalld_conf.get("Lockdown"):
                 value = self._firewalld_conf.get("Lockdown")
                 if value is not None and value.lower() in [ "yes", "true" ]:
@@ -287,20 +373,6 @@ class Firewall(object):
                 log.debug1("RFC3964_IPv4 is set to '%s'",
                            self._rfc3964_ipv4)
 
-            if self._firewalld_conf.get("AllowZoneDrifting"):
-                value = self._firewalld_conf.get("AllowZoneDrifting")
-                if value.lower() in [ "no", "false" ]:
-                    self._allow_zone_drifting = False
-                else:
-                    self._allow_zone_drifting = True
-                    if not self._offline:
-                        log.warning("AllowZoneDrifting is enabled. This is considered "
-                                    "an insecure configuration option. It will be "
-                                    "removed in a future release. Please consider "
-                                    "disabling it now.")
-                log.debug1("AllowZoneDrifting is set to '%s'",
-                           self._allow_zone_drifting)
-
         self.config.set_firewalld_conf(copy.deepcopy(self._firewalld_conf))
 
         self._select_firewall_backend(self._firewall_backend)
@@ -353,6 +425,10 @@ class Firewall(object):
             log.fatal("No zones found.")
             sys.exit(1)
 
+        # load policy files
+        self._loader(config.FIREWALLD_POLICIES, "policy")
+        self._loader(config.ETC_FIREWALLD_POLICIES, "policy")
+
         # check minimum required zones
         error = False
         for z in [ "block", "drop", "trusted" ]:
@@ -415,7 +491,7 @@ class Firewall(object):
         # only needs to be split here if there are conflicting ipset types in
         # exsting ipsets and the configuration in firewalld.
         if (reload and complete_reload) or \
-           (self.ipset_enabled and self.ipset.has_ipsets()):
+           (self.ipset.backends() and self.ipset.has_ipsets()):
             transaction.execute(True)
             transaction.clear()
 
@@ -429,7 +505,7 @@ class Firewall(object):
         transaction.clear()
 
         # apply settings for loaded ipsets while reloading here
-        if self.ipset_enabled and self.ipset.has_ipsets():
+        if (self.ipset.backends()) and self.ipset.has_ipsets():
             log.debug1("Applying ipsets")
             self.ipset.apply_ipsets()
 
@@ -446,6 +522,10 @@ class Firewall(object):
         self.zone.change_default_zone(None, self._default_zone,
                                       use_transaction=transaction)
 
+        # apply policies
+        log.debug1("Applying used policies")
+        self.policy.apply_policies(use_transaction=transaction)
+
         # Execute transaction
         transaction.execute(True)
 
@@ -606,6 +686,19 @@ class Firewall(object):
                     self.helper.add_helper(obj)
                     # add a deep copy to the configuration interface
                     self.config.add_helper(copy.deepcopy(obj))
+                elif reader_type == "policy":
+                    obj = policy_reader(filename, path)
+                    if obj.name in self.policy.get_policies():
+                        orig_obj = self.policy.get_policy(obj.name)
+                        log.debug1("  Overloads %s '%s' ('%s/%s')", reader_type,
+                                   orig_obj.name, orig_obj.path,
+                                   orig_obj.filename)
+                        self.policy.remove_policy(orig_obj.name)
+                    elif obj.path.startswith(config.ETC_FIREWALLD):
+                        obj.default = True
+                    self.policy.add_policy(obj)
+                    # add a deep copy to the configuration interface
+                    self.config.add_policy_object(copy.deepcopy(obj))
                 else:
                     log.fatal("Unknown reader type %s", reader_type)
             except FirewallError as msg:
@@ -637,15 +730,20 @@ class Firewall(object):
         self.config.cleanup()
         self.direct.cleanup()
         self.policies.cleanup()
+        self.policy.cleanup()
         self._firewalld_conf.cleanup()
         self.__init_vars()
 
     def stop(self):
-        if self.cleanup_on_exit and not self._offline:
-            self.flush()
-            self.ipset.flush()
-            self.set_policy("ACCEPT")
-            self.modules_backend.unload_firewall_modules()
+        if not self._offline:
+            if self.cleanup_on_exit:
+                self.flush()
+                self.ipset.flush()
+                self.set_policy("ACCEPT")
+
+            if self.cleanup_modules_on_exit:
+                log.debug1('Unloading firewall kernel modules')
+                self.modules_backend.unload_firewall_modules()
 
         self.cleanup()
 
@@ -879,6 +977,12 @@ class Firewall(object):
         if self._panic:
             raise FirewallError(errors.PANIC_MODE)
 
+    def check_policy(self, policy):
+        _policy = policy
+        if _policy not in self.policy.get_policies():
+            raise FirewallError(errors.INVALID_POLICY, _policy)
+        return _policy
+
     def check_zone(self, zone):
         _zone = zone
         if not _zone or _zone == "":
@@ -943,7 +1047,7 @@ class Firewall(object):
             # save zone interfaces
             _zone_interfaces = { }
             for zone in self.zone.get_zones():
-                _zone_interfaces[zone] = self.zone.get_settings(zone)["interfaces"]
+                _zone_interfaces[zone] = self.zone.get_zone(zone).interfaces
             # save direct config
             _direct_config = self.direct.get_runtime_config()
             _old_dz = self.get_default_zone()
@@ -987,8 +1091,8 @@ class Firewall(object):
                     _zone_interfaces[_new_dz] = { }
                 # default zone changed. Move interfaces from old default zone to
                 # the new one.
-                for iface, settings in list(_zone_interfaces[_old_dz].items()):
-                    if settings["__default__"]:
+                for iface in _zone_interfaces[_old_dz]:
+                    if iface in self._default_zone_interfaces:
                         # move only those that were added to default zone
                         # (not those that were added to specific zone same as
                         # default)
@@ -999,8 +1103,10 @@ class Firewall(object):
             # add interfaces to zones again
             for zone in self.zone.get_zones():
                 if zone in _zone_interfaces:
-                    self.zone.set_settings(zone, { "interfaces":
-                                                   _zone_interfaces[zone] })
+
+                    for interface_id in _zone_interfaces[zone]:
+                        self.zone.change_zone_of_interface(zone, interface_id)
+
                     del _zone_interfaces[zone]
                 else:
                     log.info1("New zone '%s'.", zone)
@@ -1106,15 +1212,57 @@ class Firewall(object):
             self._firewalld_conf.set("DefaultZone", _zone)
             self._firewalld_conf.write()
 
+            if self._offline:
+                return
+
             # remove old default zone from ZONES and add new default zone
             self.zone.change_default_zone(_old_dz, _zone)
 
             # Move interfaces from old default zone to the new one.
-            _old_dz_settings = self.zone.get_settings(_old_dz)
-            for iface, settings in list(_old_dz_settings["interfaces"].items()):
-                if settings["__default__"]:
+            for iface in self.zone.get_zone(_old_dz).interfaces:
+                if iface in self._default_zone_interfaces:
                     # move only those that were added to default zone
                     # (not those that were added to specific zone same as default)
                     self.zone.change_zone_of_interface("", iface)
         else:
             raise FirewallError(errors.ZONE_ALREADY_SET, _zone)
+
+    def combine_runtime_with_permanent_settings(self, permanent, runtime):
+        combined = permanent.copy()
+
+        for key,value in runtime.items():
+            # omit empty entries
+            if value or isinstance(value, bool):
+                combined[key] = value
+            # make sure to remove values that were in permanent, but no
+            # longer in runtime.
+            elif key in combined:
+                del combined[key]
+
+        return combined
+
+    def get_added_and_removed_settings(self, old_settings, new_settings):
+        # normalize rich rules, zones and policies use a different key
+        for rich_key in ["rich_rules", "rules_str"]:
+            if rich_key in new_settings:
+                new_settings[rich_key] = [str(Rich_Rule(rule_str=rule_str)) for rule_str in new_settings[rich_key]]
+
+        add_settings = {}
+        remove_settings = {}
+        for key in (set(old_settings.keys()) | set(new_settings.keys())):
+            if key in new_settings:
+                if isinstance(new_settings[key], list):
+                    old = set(old_settings[key] if key in old_settings else [])
+                    add_settings[key] = list(set(new_settings[key]) - old)
+                    remove_settings[key] = list((old ^ set(new_settings[key])) & old)
+                # check for bool or int because dbus.Boolean is a subclass of
+                # int (because bool can't be subclassed).
+                elif isinstance(new_settings[key], bool) or isinstance(new_settings[key], int):
+                    if not old_settings[key] and new_settings[key]:
+                        add_settings[key] = True
+                    elif old_settings[key] and not new_settings[key]:
+                        remove_settings[key] = False
+                else:
+                    raise FirewallError(errors.INVALID_SETTING, "Unhandled setting type {} key {}".format(type(new_settings[key]), key))
+
+        return (add_settings, remove_settings)
