@@ -24,7 +24,9 @@
 __all__ = [ "FirewallIPSet" ]
 
 from firewall.core.logger import log
-from firewall.core.ipset import remove_default_create_options as rm_def_cr_opts
+from firewall.core.ipset import remove_default_create_options as rm_def_cr_opts, \
+                                normalize_ipset_entry, check_entry_overlaps_existing, \
+                                check_for_overlapping_entries
 from firewall.core.io.ipset import IPSet
 from firewall import errors
 from firewall.errors import FirewallError
@@ -105,7 +107,7 @@ class FirewallIPSet(object):
                     except Exception as msg:
                         raise FirewallError(errors.COMMAND_FAILED, msg)
 
-            if self._fw.individual_calls():
+            if self._fw._individual_calls:
                 try:
                     backend.set_create(obj.name, obj.type, obj.options)
                 except Exception as msg:
@@ -117,6 +119,11 @@ class FirewallIPSet(object):
                         # no entries visible for ipsets with timeout
                         continue
 
+                try:
+                    backend.set_flush(obj.name)
+                except Exception as msg:
+                    raise FirewallError(errors.COMMAND_FAILED, msg)
+
                 for entry in obj.entries:
                     try:
                         backend.set_add(obj.name, entry)
@@ -155,8 +162,8 @@ class FirewallIPSet(object):
 
     # TYPE
 
-    def get_type(self, name):
-        return self.get_ipset(name, applied=True).type
+    def get_type(self, name, applied=True):
+        return self.get_ipset(name, applied=applied).type
 
     # DIMENSION
     def get_dimension(self, name):
@@ -173,8 +180,8 @@ class FirewallIPSet(object):
 
     # OPTIONS
 
-    def get_family(self, name):
-        obj = self.get_ipset(name, applied=True)
+    def get_family(self, name, applied=True):
+        obj = self.get_ipset(name, applied=applied)
         if "family" in obj.options:
             if obj.options["family"] == "inet6":
                 return "ipv6"
@@ -184,11 +191,13 @@ class FirewallIPSet(object):
 
     def add_entry(self, name, entry):
         obj = self.get_ipset(name, applied=True)
+        entry = normalize_ipset_entry(entry)
 
         IPSet.check_entry(entry, obj.options, obj.type)
         if entry in obj.entries:
             raise FirewallError(errors.ALREADY_ENABLED,
                                 "'%s' already is in '%s'" % (entry, name))
+        check_entry_overlaps_existing(entry, obj.entries)
 
         try:
             for backend in self.backends():
@@ -196,13 +205,13 @@ class FirewallIPSet(object):
         except Exception as msg:
             raise FirewallError(errors.COMMAND_FAILED, msg)
         else:
-            if "timeout" not in obj.options or obj.options["timeout"] == "0" \
-               and entry not in obj.entries:
+            if "timeout" not in obj.options or obj.options["timeout"] == "0":
                 # no entries visible for ipsets with timeout
                 obj.entries.append(entry)
 
     def remove_entry(self, name, entry):
         obj = self.get_ipset(name, applied=True)
+        entry = normalize_ipset_entry(entry)
 
         # no entry check for removal
         if entry not in obj.entries:
@@ -214,13 +223,13 @@ class FirewallIPSet(object):
         except Exception as msg:
             raise FirewallError(errors.COMMAND_FAILED, msg)
         else:
-            if "timeout" not in obj.options or obj.options["timeout"] == "0" \
-               and entry not in obj.entries:
+            if "timeout" not in obj.options or obj.options["timeout"] == "0":
                 # no entries visible for ipsets with timeout
                 obj.entries.remove(entry)
 
     def query_entry(self, name, entry):
         obj = self.get_ipset(name, applied=True)
+        entry = normalize_ipset_entry(entry)
         if "timeout" in obj.options and obj.options["timeout"] != "0":
             # no entries visible for ipsets with timeout
             raise FirewallError(errors.IPSET_WITH_TIMEOUT, name)
@@ -234,6 +243,8 @@ class FirewallIPSet(object):
     def set_entries(self, name, entries):
         obj = self.get_ipset(name, applied=True)
 
+        check_for_overlapping_entries(entries)
+
         for entry in entries:
             IPSet.check_entry(entry, obj.options, obj.type)
         if "timeout" not in obj.options or obj.options["timeout"] == "0":
@@ -250,7 +261,7 @@ class FirewallIPSet(object):
 
         try:
             for backend in self.backends():
-                if self._fw.individual_calls():
+                if self._fw._individual_calls:
                     for entry in obj.entries:
                         backend.set_add(obj.name, entry)
                 else:
