@@ -22,9 +22,9 @@
 __all__ = [ "Rich_Source", "Rich_Destination", "Rich_Service", "Rich_Port",
             "Rich_Protocol", "Rich_Masquerade", "Rich_IcmpBlock",
             "Rich_IcmpType",
-            "Rich_SourcePort", "Rich_ForwardPort", "Rich_Log", "Rich_Audit",
+            "Rich_SourcePort", "Rich_ForwardPort", "Rich_Log", "Rich_NFLog",
             "Rich_Accept", "Rich_Reject", "Rich_Drop", "Rich_Mark",
-            "Rich_Limit", "Rich_Rule" ]
+            "Rich_Audit", "Rich_Limit", "Rich_Rule", "Rich_Tcp_Mss_Clamp" ]
 
 from firewall import functions
 from firewall.core.ipset import check_ipset_name
@@ -63,13 +63,27 @@ class Rich_Source(object):
                                 "no address, mac and ipset")
 
 class Rich_Destination(object):
-    def __init__(self, addr, invert=False):
+    def __init__(self, addr, ipset, invert=False):
         self.addr = addr
+        if self.addr == "":
+            self.addr = None
+        self.ipset = ipset
+        if self.ipset == "":
+            self.ipset = None
         self.invert = invert
+        if self.addr is None and self.ipset is None:
+            raise FirewallError(errors.INVALID_RULE,
+                                "no address and ipset")
 
     def __str__(self):
-        return 'destination %saddress="%s"' % ("not " if self.invert else "",
-                                               self.addr)
+        ret = 'destination%s ' % (" NOT" if self.invert else "")
+        if self.addr is not None:
+            return ret + 'address="%s"' % self.addr
+        elif self.ipset is not None:
+            return ret + 'ipset="%s"' % self.ipset
+        else:
+            raise FirewallError(errors.INVALID_RULE,
+                                "no address and ipset")
 
 class Rich_Service(object):
     def __init__(self, name):
@@ -119,6 +133,13 @@ class Rich_IcmpType(object):
     def __str__(self):
         return 'icmp-type name="%s"' % (self.name)
 
+class Rich_Tcp_Mss_Clamp(object):
+    def __init__(self, value):
+        self.value = value
+
+    def __str__(self):
+        return 'tcp-mss-clamp value="%s"' % (self.value)
+
 class Rich_ForwardPort(object):
     def __init__(self, port, protocol, to_port, to_address):
         self.port = port
@@ -150,6 +171,45 @@ class Rich_Log(object):
              ' level="%s"' % (self.level) if self.level else "",
              " %s" % self.limit if self.limit else "")
 
+    def check(self):
+        if self.prefix and len(self.prefix) > 127:
+            raise FirewallError(errors.INVALID_LOG_PREFIX, "maximum accepted length of 'prefix' is 127.")
+
+        if self.level and \
+               self.level not in [ "emerg", "alert", "crit", "error",
+                                       "warning", "notice", "info", "debug" ]:
+            raise FirewallError(errors.INVALID_LOG_LEVEL, self.level)
+
+        if self.limit is not None:
+            self.limit.check()
+
+class Rich_NFLog(object):
+    def __init__(self, group=None, prefix=None, queue_size=None, limit=None):
+        self.group = group
+        self.prefix = prefix
+        self.threshold = queue_size
+        self.limit = limit
+
+    def __str__(self):
+        return 'nflog%s%s%s%s' % \
+            (' group="%s"' % (self.group) if self.group else "",
+             ' prefix="%s"' % (self.prefix) if self.prefix else "",
+             ' queue-size="%s"' % (self.threshold) if self.threshold else "",
+             " %s" % self.limit if self.limit else "")
+
+    def check(self):
+        if self.group and not functions.checkUINT16(self.group):
+            raise FirewallError(errors.INVALID_NFLOG_GROUP, "nflog 'group' must be an integer between 0 and 65535.")
+
+        if self.prefix and len(self.prefix) > 127:
+            raise FirewallError(errors.INVALID_LOG_PREFIX, "maximum accepted length of 'prefix' is 127.")
+
+        if self.threshold and not functions.checkUINT16(self.threshold):
+            raise FirewallError(errors.INVALID_NFLOG_QUEUE, "nflog 'queue-size' must be an integer between 0 and 65535.")
+
+        if self.limit is not None:
+            self.limit.check()
+
 class Rich_Audit(object):
     def __init__(self, limit=None):
         #TODO check default level in iptables
@@ -266,12 +326,16 @@ class Rich_Limit(object):
         return ''
 
 class Rich_Rule(object):
-    def __init__(self, family=None, rule_str=None):
+    priority_min = -32768
+    priority_max =  32767
+
+    def __init__(self, family=None, rule_str=None, priority=0):
         if family is not None:
             self.family = str(family)
         else:
             self.family = None
 
+        self.priority = priority
         self.source = None
         self.destination = None
         self.element = None
@@ -303,6 +367,9 @@ class Rich_Rule(object):
         if not rule_str:
             raise FirewallError(errors.INVALID_RULE, 'empty rule')
 
+        rule_str = functions.stripNonPrintableCharacters(rule_str)
+
+        self.priority = 0
         self.family = None
         self.source = None
         self.destination = None
@@ -325,17 +392,17 @@ class Rich_Rule(object):
             #print ("in_elements: ", in_elements)
             #print ("index: %s, element: %s, attribute: %s=%s" % (index, element, attr_name, attr_value))
             if attr_name:     # attribute
-                if attr_name not in ['family', 'address', 'mac', 'ipset',
+                if attr_name not in ['priority', 'family', 'address', 'mac', 'ipset',
                                      'invert', 'value',
                                      'port', 'protocol', 'to-port', 'to-addr',
-                                     'name', 'prefix', 'level', 'type',
+                                     'name', 'group', 'prefix', 'level', 'queue-size', 'type',
                                      'set']:
                     raise FirewallError(errors.INVALID_RULE, "bad attribute '%s'" % attr_name)
             else:             # element
                 if element in ['rule', 'source', 'destination', 'protocol',
                                'service', 'port', 'icmp-block', 'icmp-type', 'masquerade',
-                               'forward-port', 'source-port', 'log', 'audit',
-                               'accept', 'drop', 'reject', 'mark', 'limit', 'not', 'NOT', 'EOL']:
+                               'forward-port', 'source-port', 'log', 'nflog', 'audit',
+                               'accept', 'drop', 'reject', 'mark', 'limit', 'not', 'NOT', 'EOL', 'tcp-mss-clamp']:
                     if element == 'source' and self.source:
                         raise FirewallError(errors.INVALID_RULE, "more than one 'source' element")
                     elif element == 'destination' and self.destination:
@@ -345,8 +412,8 @@ class Rich_Rule(object):
                                      'masquerade', 'forward-port',
                                      'source-port'] and self.element:
                         raise FirewallError(errors.INVALID_RULE, "more than one element. There cannot be both '%s' and '%s' in one rule." % (element, self.element))
-                    elif element == 'log' and self.log:
-                        raise FirewallError(errors.INVALID_RULE, "more than one 'log' element")
+                    elif element in ['log', 'nflog'] and self.log:
+                        raise FirewallError(errors.INVALID_RULE, "more than one logging element")
                     elif element == 'audit' and self.audit:
                         raise FirewallError(errors.INVALID_RULE, "more than one 'audit' element")
                     elif element in ['accept', 'drop', 'reject', 'mark'] and self.action:
@@ -360,6 +427,8 @@ class Rich_Rule(object):
                 if not element and attr_name:
                     if attr_name == 'family':
                         raise FirewallError(errors.INVALID_RULE, "'family' outside of rule. Use 'rule family=...'.")
+                    elif attr_name == 'priority':
+                        raise FirewallError(errors.INVALID_RULE, "'priority' outside of rule. Use 'rule priority=...'.")
                     else:
                         raise FirewallError(errors.INVALID_RULE, "'%s' outside of any element. Use 'rule <element> %s= ...'." % (attr_name, attr_name))
                 elif 'rule' not in element:
@@ -371,6 +440,11 @@ class Rich_Rule(object):
                     if attr_value not in ['ipv4', 'ipv6']:
                         raise FirewallError(errors.INVALID_RULE, "'family' attribute cannot have '%s' value. Use 'ipv4' or 'ipv6' instead." % attr_value)
                     self.family = attr_value
+                elif attr_name == 'priority':
+                    try:
+                        self.priority = int(attr_value)
+                    except ValueError:
+                        raise FirewallError(errors.INVALID_PRIORITY, "invalid 'priority' attribute value '%s'." % attr_value)
                 elif attr_name:
                     if attr_name == 'protocol':
                         err_msg = "wrong 'protocol' usage. Use either 'rule protocol value=...' or  'rule [forward-]port protocol=...'."
@@ -385,17 +459,17 @@ class Rich_Rule(object):
                 elif element in ['not', 'NOT']:
                     attrs['invert'] = True
                 else:
-                    self.source = Rich_Source(attrs.get('address'), attrs.get('mac'), attrs.get('ipset'), attrs.get('invert'))
+                    self.source = Rich_Source(attrs.get('address'), attrs.get('mac'), attrs.get('ipset'), attrs.get('invert', False))
                     in_elements.pop() # source
                     attrs.clear()
                     index = index -1 # return token to input
             elif in_element == 'destination':
-                if attr_name in ['address', 'invert']:
+                if attr_name in ['address', 'ipset', 'invert']:
                     attrs[attr_name] = attr_value
                 elif element in ['not', 'NOT']:
                     attrs['invert'] = True
                 else:
-                    self.destination = Rich_Destination(attrs.get('address'), attrs.get('invert'))
+                    self.destination = Rich_Destination(attrs.get('address'), attrs.get('ipset'), attrs.get('invert', False))
                     in_elements.pop() # destination
                     attrs.clear()
                     index = index -1 # return token to input
@@ -405,6 +479,14 @@ class Rich_Rule(object):
                     in_elements.pop() # protocol
                 else:
                     raise FirewallError(errors.INVALID_RULE, "invalid 'protocol' element")
+            elif in_element == 'tcp-mss-clamp':
+                if attr_name == 'value':
+                    attrs[attr_name] = attr_value
+                else:
+                    self.element = Rich_Tcp_Mss_Clamp(attrs.get('value'))
+                    in_elements.pop()
+                    attrs.clear()
+                    index = index -1
             elif in_element == 'service':
                 if attr_name == 'name':
                     self.element = Rich_Service(attr_value)
@@ -462,6 +544,16 @@ class Rich_Rule(object):
                     in_elements.pop() # log
                     attrs.clear()
                     index = index -1 # return token to input
+            elif in_element == 'nflog':
+                if attr_name in ['group', 'prefix', 'queue-size']:
+                    attrs[attr_name] = attr_value
+                elif element == 'limit':
+                    in_elements.append('limit')
+                else:
+                    self.log = Rich_NFLog(attrs.get('group'), attrs.get('prefix'), attrs.get('queue-size'), attrs.get('limit'))
+                    in_elements.pop() # nflog
+                    attrs.clear()
+                    index = index -1 # return token to input
             elif in_element == 'audit':
                 if element == 'limit':
                     in_elements.append('limit')
@@ -528,15 +620,21 @@ class Rich_Rule(object):
             if type(self.element) == Rich_ForwardPort:
                 raise FirewallError(errors.MISSING_FAMILY)
 
-        if self.element is None:
+        if self.priority < self.priority_min or self.priority > self.priority_max:
+            raise FirewallError(errors.INVALID_PRIORITY, "'priority' attribute must be between %d and %d." \
+                                                         % (self.priority_min, self.priority_max))
+
+        if self.element is None and \
+           (self.log is None or (self.log is not None and self.priority == 0)):
             if self.action is None:
                 raise FirewallError(errors.INVALID_RULE, "no element, no action")
-            if self.source is None and self.destination is None:
+            if self.source is None and self.destination is None and self.priority == 0:
                 raise FirewallError(errors.INVALID_RULE, "no element, no source, no destination")
 
         if type(self.element) not in [ Rich_IcmpBlock,
                                        Rich_ForwardPort,
-                                       Rich_Masquerade ]:
+                                       Rich_Masquerade,
+                                       Rich_Tcp_Mss_Clamp ]:
             if self.log is None and self.audit is None and \
                     self.action is None:
                 raise FirewallError(errors.INVALID_RULE, "no action, no log, no audit")
@@ -568,12 +666,20 @@ class Rich_Rule(object):
 
         # destination
         if self.destination is not None:
-            if self.family is None:
-                raise FirewallError(errors.INVALID_FAMILY)
-            if self.destination.addr is None or \
-                    not functions.check_address(self.family,
-                                                self.destination.addr):
-                raise FirewallError(errors.INVALID_ADDR, str(self.destination.addr))
+            if self.destination.addr is not None:
+                if self.family is None:
+                    raise FirewallError(errors.INVALID_FAMILY)
+                if self.destination.ipset is not None:
+                    raise FirewallError(errors.INVALID_DESTINATION, "address and ipset")
+                if not functions.check_address(self.family, self.destination.addr):
+                    raise FirewallError(errors.INVALID_ADDR, str(self.destination.addr))
+
+            elif self.destination.ipset is not None:
+                if not check_ipset_name(self.destination.ipset):
+                    raise FirewallError(errors.INVALID_IPSET, str(self.destination.ipset))
+
+            else:
+                raise FirewallError(errors.INVALID_RULE, "invalid destination")
 
         # service
         if type(self.element) == Rich_Service:
@@ -644,6 +750,14 @@ class Rich_Rule(object):
             if self.element.protocol not in [ "tcp", "udp", "sctp", "dccp" ]:
                 raise FirewallError(errors.INVALID_PROTOCOL, self.element.protocol)
 
+        # tcp-mss-clamp
+        elif type(self.element) == Rich_Tcp_Mss_Clamp:
+            if self.action is not None:
+                raise FirewallError(errors.INVALID_RULE, "tcp-mss-clamp and %s are mutually exclusive" % self.action)
+            if self.element.value:
+                if not functions.checkTcpMssClamp(self.element.value):
+                    raise FirewallError(errors.INVALID_RULE, self.element.value)
+
         # other element and not empty?
         elif self.element is not None:
             raise FirewallError(errors.INVALID_RULE, "Unknown element %s" % 
@@ -651,13 +765,7 @@ class Rich_Rule(object):
 
         # log
         if self.log is not None:
-            if self.log.level and \
-               self.log.level not in [ "emerg", "alert", "crit", "error",
-                                       "warning", "notice", "info", "debug" ]:
-                raise FirewallError(errors.INVALID_LOG_LEVEL, self.log.level)
-
-            if self.log.limit is not None:
-                self.log.limit.check()
+            self.log.check()
 
         # audit
         if self.audit is not None:
@@ -679,6 +787,8 @@ class Rich_Rule(object):
 
     def __str__(self):
         ret = 'rule'
+        if self.priority:
+            ret += ' priority="%d"' % self.priority
         if self.family:
             ret += ' family="%s"' % self.family
         if self.source:
@@ -694,7 +804,7 @@ class Rich_Rule(object):
         if self.action:
             ret += " %s" % self.action
 
-        return (functions.u2b(ret)) if functions.PY2 else ret
+        return ret
 
 
 #class Rich_RawRule(object):
