@@ -88,6 +88,10 @@ INVALID_ENTRY       =  136
 INVALID_OPTION      =  137
 INVALID_HELPER      =  138
 INVALID_PRIORITY    =  139
+INVALID_POLICY      =  140
+INVALID_LOG_PREFIX  =  141
+INVALID_NFLOG_GROUP =  142
+INVALID_NFLOG_QUEUE =  143
 
 MISSING_TABLE       =  200
 MISSING_CHAIN       =  201
