@@ -19,29 +19,35 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #
 
-__all__ = [ "PY2", "getPortID", "getPortRange", "portStr", "getServiceName",
+__all__ = [ "getPortID", "getPortRange", "portStr", "getServiceName",
             "checkIP", "checkIP6", "checkIPnMask", "checkIP6nMask",
-            "checkProtocol", "checkInterface", "checkUINT32",
+            "checkProtocol", "checkInterface", "checkUINT16", "checkUINT32",
             "firewalld_is_active", "tempFile", "readfile", "writefile",
-            "enable_ip_forwarding", "get_nf_conntrack_helper_setting",
-            "set_nf_conntrack_helper_setting", "get_nf_conntrack_helpers",
-            "get_nf_nat_helpers", "check_port", "check_address",
+            "enable_ip_forwarding", "check_port", "check_address",
             "check_single_address", "check_mac", "uniqify", "ppid_of_pid",
             "max_zone_name_len", "checkUser", "checkUid", "checkCommand",
             "checkContext", "joinArgs", "splitArgs",
-            "b2u", "u2b", "u2b_if_py2" ]
+            "max_policy_name_len", "checkTcpMssClamp", 
+            "stripNonPrintableCharacters"]
 
 import socket
-import os, os.path
-import shlex, pipes
+import os
+import os.path
+import shlex
+import pipes
 import string
-import sys
 import tempfile
 from firewall.core.logger import log
-from firewall.core.prog import runProg
-from firewall.config import FIREWALLD_TEMPDIR, FIREWALLD_PIDFILE, COMMANDS
+from firewall.config import FIREWALLD_TEMPDIR, FIREWALLD_PIDFILE
 
-PY2 = sys.version < '3'
+NOPRINT_TRANS_TABLE = {
+    # Limit to C0 and C1 code points. Building entries for all unicode code
+    # points requires too much memory.
+    # C0 = [0, 31]
+    # C1 = [127, 159]
+    #
+    i: None for i in range(0, 160) if not (i > 31 and i < 127)
+}
 
 def getPortID(port):
     """ Check and Get port id from port string or port id using socket.getservbyname
@@ -73,6 +79,10 @@ def getPortRange(ports):
     @return Array containing start and end port id for a valid range or -1 if port can not be found and -2 if port is too big for integer input or -1 for invalid ranges or None if the range is ambiguous.
     """
 
+    # (port, port)  or [port, port] case
+    if isinstance(ports, tuple) or isinstance(ports, list):
+        return ports
+
     # "<port-id>" case
     if isinstance(ports, int) or ports.isdigit():
         id1 = getPortID(ports)
@@ -138,6 +148,105 @@ def portStr(port, delimiter=":"):
     else:
         return "%s%s%s" % (_range[0], delimiter, _range[1])
 
+def portInPortRange(port, range):
+    _port = getPortRange(port)
+    _range = getPortRange(range)
+
+    if len(_port) == 1:
+        if len(_range) == 1:
+            return getPortID(_port[0]) == getPortID(_range[0])
+        if len(_range) == 2 and \
+           getPortID(_port[0]) >= getPortID(_range[0]) and getPortID(_port[0]) <= getPortID(_range[1]):
+            return True
+    elif len(_port) == 2:
+        if len(_range) == 2 and \
+           getPortID(_port[0]) >= getPortID(_range[0]) and getPortID(_port[0]) <= getPortID(_range[1]) and \
+           getPortID(_port[1]) >= getPortID(_range[0]) and getPortID(_port[1]) <= getPortID(_range[1]):
+            return True
+
+    return False
+
+def coalescePortRange(new_range, ranges):
+    """ Coalesce a port range with existing list of port ranges
+
+        @param new_range tuple/list/string
+        @param ranges list of tuple/list/string
+        @return tuple of (list of ranges added after coalescing, list of removed original ranges)
+    """
+
+    coalesced_range = getPortRange(new_range)
+    # normalize singleton ranges, e.g. (x,) --> (x,x)
+    if len(coalesced_range) == 1:
+        coalesced_range = (coalesced_range[0], coalesced_range[0])
+    _ranges = map(getPortRange, ranges)
+    _ranges = sorted(map(lambda x: (x[0],x[0]) if len(x) == 1 else x, _ranges), key=lambda x: x[0])
+
+    removed_ranges = []
+    for range in _ranges:
+        if coalesced_range[0] <= range[0] and coalesced_range[1] >= range[1]:
+            # new range covers this
+            removed_ranges.append(range)
+        elif coalesced_range[0] <= range[0] and coalesced_range[1] <  range[1] and \
+                                                coalesced_range[1] >= range[0]:
+            # expand beginning of range
+            removed_ranges.append(range)
+            coalesced_range = (coalesced_range[0], range[1])
+        elif coalesced_range[0] >  range[0] and coalesced_range[1] >= range[1] and \
+                                                coalesced_range[0] <= range[1]:
+            # expand end of range
+            removed_ranges.append(range)
+            coalesced_range = (range[0], coalesced_range[1])
+
+    # normalize singleton ranges, e.g. (x,x) --> (x,)
+    removed_ranges = list(map(lambda x: (x[0],) if x[0] == x[1] else x, removed_ranges))
+    if coalesced_range[0] == coalesced_range[1]:
+        coalesced_range = (coalesced_range[0],)
+
+    return ([coalesced_range], removed_ranges)
+
+def breakPortRange(remove_range, ranges):
+    """ break a port range from existing list of port ranges
+
+        @param remove_range tuple/list/string
+        @param ranges list of tuple/list/string
+        @return tuple of (list of ranges added after breaking up, list of removed original ranges)
+    """
+
+    remove_range = getPortRange(remove_range)
+    # normalize singleton ranges, e.g. (x,) --> (x,x)
+    if len(remove_range) == 1:
+        remove_range = (remove_range[0], remove_range[0])
+    _ranges = map(getPortRange, ranges)
+    _ranges = sorted(map(lambda x: (x[0],x[0]) if len(x) == 1 else x, _ranges), key=lambda x: x[0])
+
+    removed_ranges = []
+    added_ranges = []
+    for range in _ranges:
+        if remove_range[0] <= range[0] and remove_range[1] >= range[1]:
+            # remove entire range
+            removed_ranges.append(range)
+        elif remove_range[0] <= range[0] and remove_range[1] <  range[1] and \
+                                             remove_range[1] >= range[0]:
+            # remove from beginning of range
+            removed_ranges.append(range)
+            added_ranges.append((remove_range[1] + 1, range[1]))
+        elif remove_range[0] >  range[0] and remove_range[1] >= range[1] and \
+                                             remove_range[0] <= range[1]:
+            # remove from end of range
+            removed_ranges.append(range)
+            added_ranges.append((range[0], remove_range[0] - 1))
+        elif remove_range[0] > range[0] and remove_range[1] < range[1]:
+            # remove inside range
+            removed_ranges.append(range)
+            added_ranges.append((range[0], remove_range[0] - 1))
+            added_ranges.append((remove_range[1] + 1, range[1]))
+
+    # normalize singleton ranges, e.g. (x,x) --> (x,)
+    removed_ranges = list(map(lambda x: (x[0],) if x[0] == x[1] else x, removed_ranges))
+    added_ranges = list(map(lambda x: (x[0],) if x[0] == x[1] else x, added_ranges))
+
+    return (added_ranges, removed_ranges)
+
 def getServiceName(port, proto):
     """ Check and Get service name from port and proto string combination using socket.getservbyport
 
@@ -165,6 +274,14 @@ def checkIP(ip):
         return False
     return True
 
+def normalizeIP6(ip):
+    """ Normalize the IPv6 address
+
+    This is mostly about converting URL-like IPv6 address to normal ones.
+    e.g. [1234::4321] --> 1234:4321
+    """
+    return ip.strip("[]")
+
 def checkIP6(ip):
     """ Check IPv6 address.
     
@@ -173,7 +290,7 @@ def checkIP6(ip):
     """
 
     try:
-        socket.inet_pton(socket.AF_INET6, ip)
+        socket.inet_pton(socket.AF_INET6, normalizeIP6(ip))
     except socket.error:
         return False
     return True
@@ -201,6 +318,9 @@ def checkIPnMask(ip):
                 return False
     return True
 
+def stripNonPrintableCharacters(rule_str):
+    return rule_str.translate(NOPRINT_TRANS_TABLE)
+
 def checkIP6nMask(ip):
     if "/" in ip:
         addr = ip[:ip.index("/")]
@@ -237,6 +357,17 @@ def checkProtocol(protocol):
 
     return True
 
+def checkTcpMssClamp(tcp_mss_clamp_value):
+    if tcp_mss_clamp_value:
+        if tcp_mss_clamp_value.isdigit():
+            if int(tcp_mss_clamp_value) < 536:
+                return False
+        elif tcp_mss_clamp_value == "None":
+            return True
+        elif tcp_mss_clamp_value != "pmtu":
+            return False
+    return True
+
 def checkInterface(iface):
     """ Check interface string
 
@@ -256,6 +387,16 @@ def checkInterface(iface):
     #    return False
     return True
 
+def checkUINT16(val):
+    try:
+        x = int(val, 0)
+    except ValueError:
+        return False
+    else:
+        if x >= 0 and x <= 65535:
+            return True
+    return False
+
 def checkUINT32(val):
     try:
         x = int(val, 0)
@@ -331,56 +472,8 @@ def enable_ip_forwarding(ipv):
         return writefile("/proc/sys/net/ipv6/conf/all/forwarding", "1\n")
     return False
 
-def get_nf_conntrack_helpers():
-    kver = os.uname()[2]
-    path = "/lib/modules/%s/kernel/net/netfilter/" % kver
-    helpers = { }
-    if os.path.isdir(path):
-        for filename in sorted(os.listdir(path)):
-            if not filename.startswith("nf_conntrack_"):
-                continue
-            module = filename.split(".")[0]
-            (status, ret) = runProg(COMMANDS["modinfo"], [ module, ])
-            if status != 0:
-                continue
-            for line in ret.split("\n"):
-                if line.startswith("alias:") and "-helper-" in line:
-                    helper = line.split(":")[1].strip()
-                    helper = helper.replace("nfct-helper-", "")
-                    helper = helper.replace("_", "-")
-                    helpers.setdefault(module, [ ]).append(helper)
-    return helpers
-
-def get_nf_nat_helpers():
-    kver = os.uname()[2]
-    path = "/lib/modules/%s/kernel/net/netfilter/" % kver
-    helpers = { }
-    if os.path.isdir(path):
-        for filename in sorted(os.listdir(path)):
-            if not filename.startswith("nf_nat_"):
-                continue
-            module = filename.split(".")[0]
-            (status, ret) = runProg(COMMANDS["modinfo"], [ module, ])
-            if status != 0:
-                continue
-            alias = None
-            for line in ret.split("\n"):
-                if line.startswith("description:") and "NAT helper" in line:
-                    helper = module.replace("nf_nat_", "")
-                    helper = helper.replace("_", "-")
-                    helpers.setdefault(module, [ ]).append(helper)
-    return helpers
-
-def get_nf_conntrack_helper_setting():
-    try:
-        return int(readfile("/proc/sys/net/netfilter/nf_conntrack_helper")[0])
-    except Exception:
-        log.warning("Failed to get and parse nf_conntrack_helper setting")
-        return 0
-
-def set_nf_conntrack_helper_setting(flag):
-    return writefile("/proc/sys/net/netfilter/nf_conntrack_helper",
-                     "1\n" if flag else "0\n")
+def get_nf_conntrack_short_name(module):
+    return module.replace("_","-").replace("nf-conntrack-", "")
 
 def check_port(port):
     _range = getPortRange(port)
@@ -443,10 +536,21 @@ def ppid_of_pid(pid):
         return None
     return pid
 
+def max_policy_name_len():
+    """
+    iptables limits length of chain to (currently) 28 chars.
+    The longest chain we create is POST_<policy>_allow,
+    which leaves 28 - 11 = 17 chars for <policy>.
+    """
+    from firewall.core.ipXtables import POLICY_CHAIN_PREFIX
+    from firewall.core.base import SHORTCUTS
+    longest_shortcut = max(map(len, SHORTCUTS.values()))
+    return 28 - (longest_shortcut + len(POLICY_CHAIN_PREFIX) + len("_allow"))
+
 def max_zone_name_len():
     """
     Netfilter limits length of chain to (currently) 28 chars.
-    The longest chain we create is FWDI_<zone>_allow,
+    The longest chain we create is POST_<zone>_allow,
     which leaves 28 - 11 = 17 chars for <zone>.
     """
     from firewall.core.base import SHORTCUTS
@@ -508,28 +612,4 @@ def joinArgs(args):
         return " ".join(pipes.quote(a) for a in args)
 
 def splitArgs(_string):
-    if PY2 and isinstance(_string, unicode):
-        # Python2's shlex doesn't like unicode
-        _string = u2b(_string)
-        splits = shlex.split(_string)
-        return map(b2u, splits)
-    else:
-        return shlex.split(_string)
-
-def b2u(_string):
-    """ bytes to unicode """
-    if isinstance(_string, bytes):
-        return _string.decode('UTF-8', 'replace')
-    return _string
-
-def u2b(_string):
-    """ unicode to bytes """
-    if not isinstance(_string, bytes):
-        return _string.encode('UTF-8', 'replace')
-    return _string
-
-def u2b_if_py2(_string):
-    """ unicode to bytes only if Python 2"""
-    if PY2 and isinstance(_string, unicode):
-        return _string.encode('UTF-8', 'replace')
-    return _string
+    return shlex.split(_string)
