@@ -18,29 +18,26 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-# force use of pygobject3 in python-slip
-from gi.repository import GObject
-import sys
-sys.modules['gobject'] = GObject
 import os
 
 import dbus
 import dbus.service
-import slip.dbus
-import slip.dbus.service
 
 from firewall import config
 from firewall.core.base import DEFAULT_ZONE_TARGET
 from firewall.core.watcher import Watcher
 from firewall.core.logger import log
+from firewall.server.dbus import DbusServiceObject
 from firewall.server.decorators import handle_exceptions, \
-    dbus_handle_exceptions, dbus_service_method
+    dbus_handle_exceptions, dbus_service_method, \
+    dbus_service_method_deprecated, dbus_service_signal_deprecated, \
+    dbus_polkit_require_auth
 from firewall.server.config_icmptype import FirewallDConfigIcmpType
 from firewall.server.config_service import FirewallDConfigService
 from firewall.server.config_zone import FirewallDConfigZone
+from firewall.server.config_policy import FirewallDConfigPolicy
 from firewall.server.config_ipset import FirewallDConfigIPSet
 from firewall.server.config_helper import FirewallDConfigHelper
-from firewall.core.io.zone import Zone
 from firewall.core.io.icmptype import IcmpType
 from firewall.core.io.ipset import IPSet
 from firewall.core.io.helper import Helper
@@ -49,7 +46,8 @@ from firewall.core.io.direct import Direct
 from firewall.dbus_utils import dbus_to_python, \
     command_of_sender, context_of_sender, uid_of_sender, user_of_uid, \
     dbus_introspection_prepare_properties, \
-    dbus_introspection_add_properties
+    dbus_introspection_add_properties, \
+    dbus_introspection_add_deprecated
 from firewall import errors
 from firewall.errors import FirewallError
 
@@ -59,7 +57,7 @@ from firewall.errors import FirewallError
 #
 ############################################################################
 
-class FirewallDConfig(slip.dbus.service.Object):
+class FirewallDConfig(DbusServiceObject):
     """FirewallD main class"""
 
     persistent = True
@@ -85,6 +83,8 @@ class FirewallDConfig(slip.dbus.service.Object):
         self.watcher.add_watch_dir(config.ETC_FIREWALLD_SERVICES)
         self.watcher.add_watch_dir(config.FIREWALLD_ZONES)
         self.watcher.add_watch_dir(config.ETC_FIREWALLD_ZONES)
+        self.watcher.add_watch_dir(config.FIREWALLD_POLICIES)
+        self.watcher.add_watch_dir(config.ETC_FIREWALLD_POLICIES)
         # Add watches for combined zone directories
         if os.path.exists(config.ETC_FIREWALLD_ZONES):
             for filename in sorted(os.listdir(config.ETC_FIREWALLD_ZONES)):
@@ -98,6 +98,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         dbus_introspection_prepare_properties(self,
                                               config.dbus.DBUS_INTERFACE_CONFIG,
                                               { "CleanupOnExit": "readwrite",
+                                                "CleanupModulesOnExit": "readwrite",
                                                 "IPv6_rpfilter": "readwrite",
                                                 "Lockdown": "readwrite",
                                                 "MinimalMark": "readwrite",
@@ -122,6 +123,8 @@ class FirewallDConfig(slip.dbus.service.Object):
         self.zone_idx = 0
         self.helpers = [ ]
         self.helper_idx = 0
+        self.policy_objects = [ ]
+        self.policy_object_idx = 0
 
         for ipset in self.config.get_ipsets():
             self._addIPSet(self.config.get_ipset(ipset))
@@ -133,6 +136,8 @@ class FirewallDConfig(slip.dbus.service.Object):
             self._addZone(self.config.get_zone(zone))
         for helper in self.config.get_helpers():
             self._addHelper(self.config.get_helper(helper))
+        for policy in self.config.get_policy_objects():
+            self._addPolicy(self.config.get_policy_object(policy))
 
     @handle_exceptions
     def __del__(self):
@@ -160,6 +165,10 @@ class FirewallDConfig(slip.dbus.service.Object):
             item = self.helpers.pop()
             item.unregister()
             del item
+        while len(self.policy_objects) > 0:
+            item = self.policy_objects.pop()
+            item.unregister()
+            del item
         self._init_vars()
 
     @handle_exceptions
@@ -292,6 +301,20 @@ class FirewallDConfig(slip.dbus.service.Object):
                 return
             self.Updated()
 
+        elif (name.startswith(config.FIREWALLD_POLICIES) or \
+              name.startswith(config.ETC_FIREWALLD_POLICIES)) and \
+             name.endswith(".xml"):
+            try:
+                (what, obj) = self.config.update_policy_object_from_path(name)
+            except Exception as msg:
+                log.error("Failed to load policy file '%s': %s" % (name, msg))
+                return
+            if what == "new":
+                self._addPolicy(obj)
+            elif what == "remove":
+                self.removePolicy(obj)
+            elif what == "update":
+                self._updatePolicy(obj)
 
     @handle_exceptions
     def _addIcmpType(self, obj):
@@ -325,6 +348,14 @@ class FirewallDConfig(slip.dbus.service.Object):
                 zone.obj = self.config.set_zone_config(zone.obj, settings)
                 zone.Updated(zone.obj.name)
 
+        for policy in self.policy_objects:
+            settings = policy.getSettings()
+            # if this IcmpType is used in a policy remove it from that policy first
+            if "icmp_blocks" in settings and obj.name in settings["icmp_blocks"]:
+                settings["icmp_blocks"].remove(obj.name)
+                policy.obj = self.config.set_policy_object_config_dict(policy.obj, settings)
+                policy.Updated(policy.obj.name)
+
         for icmptype in self.icmptypes:
             if icmptype.obj == obj:
                 icmptype.Removed(obj.name)
@@ -363,6 +394,14 @@ class FirewallDConfig(slip.dbus.service.Object):
                 zone.obj = self.config.set_zone_config(zone.obj, settings)
                 zone.Updated(zone.obj.name)
 
+        for policy in self.policy_objects:
+            settings = policy.getSettings()
+            # if this Service is used in a policy remove it from that policy first
+            if "services" in settings and obj.name in settings["services"]:
+                settings["services"].remove(obj.name)
+                policy.obj = self.config.set_policy_object_config_dict(policy.obj, settings)
+                policy.Updated(policy.obj.name)
+
         for service in self.services:
             if service.obj == obj:
                 service.Removed(obj.name)
@@ -398,6 +437,34 @@ class FirewallDConfig(slip.dbus.service.Object):
                 self.zones.remove(zone)
                 del zone
 
+    @handle_exceptions
+    def _addPolicy(self, obj):
+        # TODO: check for idx overflow
+        config_policy = FirewallDConfigPolicy(
+            self, self.config, obj, self.policy_object_idx, self.busname,
+            "%s/%d" % (config.dbus.DBUS_PATH_CONFIG_POLICY, self.policy_object_idx))
+        self.policy_objects.append(config_policy)
+        self.policy_object_idx += 1
+        self.PolicyAdded(obj.name)
+        return config_policy
+
+    @handle_exceptions
+    def _updatePolicy(self, obj):
+        for policy in self.policy_objects:
+            if policy.obj.name == obj.name and policy.obj.path == obj.path and \
+                    policy.obj.filename == obj.filename:
+                policy.obj = obj
+                policy.Updated(obj.name)
+
+    @handle_exceptions
+    def removePolicy(self, obj):
+        for policy in self.policy_objects:
+            if policy.obj == obj:
+                policy.Removed(obj.name)
+                policy.unregister()
+                self.policy_objects.remove(policy)
+                del policy
+
     @handle_exceptions
     def _addIPSet(self, obj):
         # TODO: check for idx overflow
@@ -486,9 +553,9 @@ class FirewallDConfig(slip.dbus.service.Object):
     @dbus_handle_exceptions
     def _get_property(self, prop):
         if prop not in [ "DefaultZone", "MinimalMark", "CleanupOnExit",
-                         "Lockdown", "IPv6_rpfilter", "IndividualCalls",
-                         "LogDenied", "AutomaticHelpers", "FirewallBackend",
-                         "FlushAllOnReload", "RFC3964_IPv4",
+                         "CleanupModulesOnExit", "Lockdown", "IPv6_rpfilter",
+                         "IndividualCalls", "LogDenied", "AutomaticHelpers",
+                         "FirewallBackend", "FlushAllOnReload", "RFC3964_IPv4",
                          "AllowZoneDrifting" ]:
             raise dbus.exceptions.DBusException(
                 "org.freedesktop.DBus.Error.InvalidArgs: "
@@ -510,6 +577,10 @@ class FirewallDConfig(slip.dbus.service.Object):
             if value is None:
                 value = "yes" if config.FALLBACK_CLEANUP_ON_EXIT else "no"
             return dbus.String(value)
+        elif prop == "CleanupModulesOnExit":
+            if value is None:
+                value = "yes" if config.FALLBACK_CLEANUP_MODULES_ON_EXIT else "no"
+            return dbus.String(value)
         elif prop == "Lockdown":
             if value is None:
                 value = "yes" if config.FALLBACK_LOCKDOWN else "no"
@@ -555,6 +626,8 @@ class FirewallDConfig(slip.dbus.service.Object):
             return dbus.Int32(self._get_property(prop))
         elif prop == "CleanupOnExit":
             return dbus.String(self._get_property(prop))
+        elif prop == "CleanupModulesOnExit":
+            return dbus.String(self._get_property(prop))
         elif prop == "Lockdown":
             return dbus.String(self._get_property(prop))
         elif prop == "IPv6_rpfilter":
@@ -599,8 +672,6 @@ class FirewallDConfig(slip.dbus.service.Object):
                 "org.freedesktop.DBus.Error.UnknownInterface: "
                 "Interface '%s' does not exist" % interface_name)
 
-        return self._get_dbus_property(property_name)
-
     @dbus_service_method(dbus.PROPERTIES_IFACE, in_signature='s',
                          out_signature='a{sv}')
     @dbus_handle_exceptions
@@ -611,9 +682,9 @@ class FirewallDConfig(slip.dbus.service.Object):
         ret = { }
         if interface_name == config.dbus.DBUS_INTERFACE_CONFIG:
             for x in [ "DefaultZone", "MinimalMark", "CleanupOnExit",
-                       "Lockdown", "IPv6_rpfilter", "IndividualCalls",
-                       "LogDenied", "AutomaticHelpers", "FirewallBackend",
-                       "FlushAllOnReload", "RFC3964_IPv4",
+                       "CleanupModulesOnExit", "Lockdown", "IPv6_rpfilter",
+                       "IndividualCalls", "LogDenied", "AutomaticHelpers",
+                       "FirewallBackend", "FlushAllOnReload", "RFC3964_IPv4",
                        "AllowZoneDrifting" ]:
                 ret[x] = self._get_property(x)
         elif interface_name in [ config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
@@ -626,7 +697,7 @@ class FirewallDConfig(slip.dbus.service.Object):
 
         return dbus.Dictionary(ret, signature="sv")
 
-    @slip.dbus.polkit.require_auth(config.dbus.PK_ACTION_CONFIG)
+    @dbus_polkit_require_auth(config.dbus.PK_ACTION_CONFIG)
     @dbus_service_method(dbus.PROPERTIES_IFACE, in_signature='ssv')
     @dbus_handle_exceptions
     def Set(self, interface_name, property_name, new_value, sender=None):
@@ -638,64 +709,43 @@ class FirewallDConfig(slip.dbus.service.Object):
         self.accessCheck(sender)
 
         if interface_name == config.dbus.DBUS_INTERFACE_CONFIG:
-            if property_name in [ "MinimalMark", "CleanupOnExit", "Lockdown",
+            if property_name in [ "CleanupOnExit",
+                                  "CleanupModulesOnExit", "Lockdown",
                                   "IPv6_rpfilter", "IndividualCalls",
-                                  "LogDenied", "AutomaticHelpers",
+                                  "LogDenied",
                                   "FirewallBackend", "FlushAllOnReload",
-                                  "RFC3964_IPv4", "AllowZoneDrifting" ]:
-                if property_name == "MinimalMark":
-                    try:
-                        int(new_value)
-                    except ValueError:
-                        raise FirewallError(errors.INVALID_MARK, new_value)
-                try:
-                    new_value = str(new_value)
-                except:
-                    raise FirewallError(errors.INVALID_VALUE,
-                                        "'%s' for %s" % \
-                                        (new_value, property_name))
-                if property_name in [ "CleanupOnExit", "Lockdown",
-                                      "IPv6_rpfilter", "IndividualCalls" ]:
+                                  "RFC3964_IPv4"]:
+                if property_name in [ "CleanupOnExit", "CleanupModulesOnExit",
+                                      "Lockdown", "IPv6_rpfilter",
+                                      "IndividualCalls", "FlushAllOnReload",
+                                      "RFC3964_IPv4"]:
                     if new_value.lower() not in [ "yes", "no",
                                                   "true", "false" ]:
                         raise FirewallError(errors.INVALID_VALUE,
                                             "'%s' for %s" % \
                                             (new_value, property_name))
-                if property_name == "LogDenied":
+                elif property_name == "LogDenied":
                     if new_value not in config.LOG_DENIED_VALUES:
                         raise FirewallError(errors.INVALID_VALUE,
                                             "'%s' for %s" % \
                                             (new_value, property_name))
-                if property_name == "AutomaticHelpers":
-                    if new_value not in config.AUTOMATIC_HELPERS_VALUES:
-                        raise FirewallError(errors.INVALID_VALUE,
-                                            "'%s' for %s" % \
-                                            (new_value, property_name))
-                if property_name == "FirewallBackend":
+                elif property_name == "FirewallBackend":
                     if new_value not in config.FIREWALL_BACKEND_VALUES:
                         raise FirewallError(errors.INVALID_VALUE,
                                             "'%s' for %s" % \
                                             (new_value, property_name))
-                if property_name == "FlushAllOnReload":
-                    if new_value.lower() not in ["yes", "true", "no", "false"]:
-                        raise FirewallError(errors.INVALID_VALUE,
-                                            "'%s' for %s" % \
-                                            (new_value, property_name))
-                if property_name == "RFC3964_IPv4":
-                    if new_value.lower() not in ["yes", "true", "no", "false"]:
-                        raise FirewallError(errors.INVALID_VALUE,
-                                            "'%s' for %s" % \
-                                            (new_value, property_name))
-                if property_name == "AllowZoneDrifting":
-                    if new_value.lower() not in ["yes", "true", "no", "false"]:
-                        raise FirewallError(errors.INVALID_VALUE,
-                                            "'%s' for %s" % \
-                                            (new_value, property_name))
+                else:
+                    raise dbus.exceptions.DBusException(
+                        "org.freedesktop.DBus.Error.InvalidArgs: "
+                        "Property '%s' does not exist" % property_name)
 
                 self.config.get_firewalld_conf().set(property_name, new_value)
                 self.config.get_firewalld_conf().write()
                 self.PropertiesChanged(interface_name,
                                        { property_name: new_value }, [ ])
+            elif property_name in ["MinimalMark", "AutomaticHelpers", "AllowZoneDrifting"]:
+                # deprecated fields. Ignore setting them.
+                pass
             else:
                 raise dbus.exceptions.DBusException(
                     "org.freedesktop.DBus.Error.InvalidArgs: "
@@ -719,7 +769,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         log.debug1("config.PropertiesChanged('%s', '%s', '%s')",
                    interface_name, changed_properties, invalidated_properties)
 
-    @slip.dbus.polkit.require_auth(config.dbus.PK_ACTION_INFO)
+    @dbus_polkit_require_auth(config.dbus.PK_ACTION_INFO)
     @dbus_service_method(dbus.INTROSPECTABLE_IFACE, out_signature='s')
     @dbus_handle_exceptions
     def Introspect(self, sender=None): # pylint: disable=W0613
@@ -727,8 +777,16 @@ class FirewallDConfig(slip.dbus.service.Object):
 
         data = super(FirewallDConfig, self).Introspect(self.path,
                                                        self.busname.get_bus())
-        return dbus_introspection_add_properties(
-            self, data, config.dbus.DBUS_INTERFACE_CONFIG)
+        data = dbus_introspection_add_properties(
+                    self, data, config.dbus.DBUS_INTERFACE_CONFIG)
+
+        for interface in [config.dbus.DBUS_INTERFACE_CONFIG_DIRECT]:
+            data = dbus_introspection_add_deprecated(
+                        self, data, interface,
+                        dbus_service_method_deprecated().deprecated,
+                        dbus_service_signal_deprecated().deprecated)
+
+        return data
 
     # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
 
@@ -747,7 +805,7 @@ class FirewallDConfig(slip.dbus.service.Object):
     def setLockdownWhitelist(self, settings, sender=None): # pylint: disable=W0613
         log.debug1("config.policies.setLockdownWhitelist(...)")
         settings = dbus_to_python(settings)
-        self.config.get_policies().lockdown_whitelist.import_config(settings)
+        self.config.get_policies().lockdown_whitelist.import_config(settings, {})
         self.config.get_policies().lockdown_whitelist.write()
         self.LockdownWhitelistUpdated()
 
@@ -1187,7 +1245,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         return ret[0] if ret else ""
 
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG,
-                         in_signature='s'+Zone.DBUS_SIGNATURE,
+                         in_signature="s(sssbsasa(ss)asba(ssss)asasasasa(ss)b)",
                          out_signature='o')
     @dbus_handle_exceptions
     def addZone(self, zone, settings, sender=None):
@@ -1206,11 +1264,82 @@ class FirewallDConfig(slip.dbus.service.Object):
         config_zone = self._addZone(obj)
         return config_zone
 
+    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG,
+                         in_signature="sa{sv}",
+                         out_signature='o')
+    @dbus_handle_exceptions
+    def addZone2(self, zone, settings, sender=None):
+        """add zone with given name and settings
+        """
+        zone = dbus_to_python(zone, str)
+        settings = dbus_to_python(settings)
+        log.debug1("config.addZone('%s')", zone)
+        self.accessCheck(sender)
+        if "target" in settings and settings["target"] == "default":
+            settings["target"] = DEFAULT_ZONE_TARGET
+        obj = self.config.new_zone_dict(zone, settings)
+        config_zone = self._addZone(obj)
+        return config_zone
+
     @dbus.service.signal(config.dbus.DBUS_INTERFACE_CONFIG, signature='s')
     @dbus_handle_exceptions
     def ZoneAdded(self, zone):
         log.debug1("config.ZoneAdded('%s')" % (zone))
 
+    # policies
+
+    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG, out_signature='ao')
+    @dbus_handle_exceptions
+    def listPolicies(self, sender=None):
+        """list policies objects paths
+        """
+        log.debug1("config.listPolicies()")
+        return self.policy_objects
+
+    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG, out_signature='as')
+    @dbus_handle_exceptions
+    def getPolicyNames(self, sender=None):
+        """get policy names
+        """
+        log.debug1("config.getPolicyNames()")
+        policies = [ ]
+        for obj in self.policy_objects:
+            policies.append(obj.obj.name)
+        return sorted(policies)
+
+    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG, in_signature='s',
+                         out_signature='o')
+    @dbus_handle_exceptions
+    def getPolicyByName(self, policy, sender=None):
+        """object path of policy with given name
+        """
+        policy = dbus_to_python(policy, str)
+        log.debug1("config.getPolicyByName('%s')", policy)
+        for obj in self.policy_objects:
+            if obj.obj.name == policy:
+                return obj
+        raise FirewallError(errors.INVALID_POLICY, policy)
+
+    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG,
+                         in_signature="sa{sv}",
+                         out_signature='o')
+    @dbus_handle_exceptions
+    def addPolicy(self, policy, settings, sender=None):
+        """add policy with given name and settings
+        """
+        policy = dbus_to_python(policy, str)
+        settings = dbus_to_python(settings)
+        log.debug1("config.addPolicy('%s')", policy)
+        self.accessCheck(sender)
+        obj = self.config.new_policy_object_dict(policy, settings)
+        config_policy = self._addPolicy(obj)
+        return config_policy
+
+    @dbus.service.signal(config.dbus.DBUS_INTERFACE_CONFIG, signature='s')
+    @dbus_handle_exceptions
+    def PolicyAdded(self, policy):
+        log.debug1("config.PolicyAdded('%s')" % (policy))
+
     # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
 
     # H E L P E R S
@@ -1271,6 +1400,7 @@ class FirewallDConfig(slip.dbus.service.Object):
     # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
     # DIRECT
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          out_signature=Direct.DBUS_SIGNATURE)
     @dbus_handle_exceptions
@@ -1279,6 +1409,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         log.debug1("config.direct.getSettings()")
         return self.config.get_direct().export_config()
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature=Direct.DBUS_SIGNATURE)
     @dbus_handle_exceptions
@@ -1286,10 +1417,11 @@ class FirewallDConfig(slip.dbus.service.Object):
         # returns list ipv, table, list of chains
         log.debug1("config.direct.update()")
         settings = dbus_to_python(settings)
-        self.config.get_direct().import_config(settings)
+        self.config.get_direct().import_config(settings, {})
         self.config.get_direct().write()
         self.Updated()
 
+    @dbus_service_signal_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus.service.signal(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_handle_exceptions
     def Updated(self):
@@ -1297,6 +1429,7 @@ class FirewallDConfig(slip.dbus.service.Object):
 
     # chain
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sss')
     @dbus_handle_exceptions
@@ -1316,6 +1449,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         settings[0].append(idx)
         self.update(settings)
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sss')
     @dbus_handle_exceptions
@@ -1335,6 +1469,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         settings[0].remove(idx)
         self.update(settings)
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sss', out_signature='b')
     @dbus_handle_exceptions
@@ -1347,6 +1482,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         idx = tuple((ipv, table, chain))
         return idx in self.getSettings()[0]
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='ss', out_signature='as')
     @dbus_handle_exceptions
@@ -1360,6 +1496,7 @@ class FirewallDConfig(slip.dbus.service.Object):
                 ret.append(idx[2])
         return ret
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='', out_signature='a(sss)')
     @dbus_handle_exceptions
@@ -1369,6 +1506,7 @@ class FirewallDConfig(slip.dbus.service.Object):
 
     # rule
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sssias')
     @dbus_handle_exceptions
@@ -1390,6 +1528,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         settings[1].append(idx)
         self.update(tuple(settings))
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sssias')
     @dbus_handle_exceptions
@@ -1411,6 +1550,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         settings[1].remove(idx)
         self.update(tuple(settings))
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sssias', out_signature='b')
     @dbus_handle_exceptions
@@ -1425,6 +1565,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         idx = (ipv, table, chain, priority, args)
         return idx in self.getSettings()[1]
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sss')
     @dbus_handle_exceptions
@@ -1441,6 +1582,7 @@ class FirewallDConfig(slip.dbus.service.Object):
                 settings[1].remove(rule)
         self.update(tuple(settings))
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sss', out_signature='a(ias)')
     @dbus_handle_exceptions
@@ -1456,6 +1598,7 @@ class FirewallDConfig(slip.dbus.service.Object):
                 ret.append((idx[3], idx[4]))
         return ret
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='', out_signature='a(sssias)')
     @dbus_handle_exceptions
@@ -1465,6 +1608,7 @@ class FirewallDConfig(slip.dbus.service.Object):
 
     # passthrough
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sas')
     @dbus_handle_exceptions
@@ -1483,6 +1627,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         self.update(settings)
 
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sas')
     @dbus_handle_exceptions
@@ -1500,6 +1645,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         settings[2].remove(idx)
         self.update(settings)
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sas', out_signature='b')
     @dbus_handle_exceptions
@@ -1511,6 +1657,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         idx = (ipv, args)
         return idx in self.getSettings()[2]
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='s', out_signature='aas')
     @dbus_handle_exceptions
@@ -1523,6 +1670,7 @@ class FirewallDConfig(slip.dbus.service.Object):
                 ret.append(idx[1])
         return ret
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          out_signature='a(sas)')
     @dbus_handle_exceptions
